
#ifndef MGA_COREBINFILE_H
#define MGA_COREBINFILE_H

#include <fstream.h>
#include <slist>
#include <hash_map>

class CCoreBinFile;

// --------------------------- BinAttr

class BinAttrBase
{
public:
	BinAttrBase() : attrid(ATTRID_NONE) { }
	virtual ~BinAttrBase() { }

	attrid_type attrid;

	static BinAttrBase *Create(valtype_type valtype);

	virtual valtype_type GetValType() const NOTHROW = 0;
	virtual void Set(CCoreBinFile *binfile, VARIANT p) = 0;
	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const = 0;
	virtual void Write(CCoreBinFile *binfile) const = 0;
	virtual void Read(CCoreBinFile *binfile) = 0;
};

typedef slist<BinAttrBase*> binattrs_type;
typedef binattrs_type::iterator binattrs_iterator;

template<int VALTYPE>
class BinAttr;

// --------------------------- BinObject

class BinObject
{
public:
	~BinObject() { DestroyAttributes(); }

	binattrs_type binattrs;
	bool deleted;

	BinAttrBase *Find(attrid_type attrid)
	{
		binattrs_iterator i = binattrs.begin();
		binattrs_iterator e = binattrs.end();
		while( i != e && (*i)->attrid != attrid )
			++i;

		if( i == e )
			HR_THROW(E_BINFILE);

		return *i;
	}

	void CreateAttributes(ICoreMetaObject *metaobject);
	void DestroyAttributes();
	void Read(CCoreBinFile *binfile);
	void Write(CCoreBinFile *binfile);

	bool HasEmptyPointers() const;
};

typedef hash_map<metaobjidpair_type, BinObject, 
	metaobjidpair_hashfunc, metaobjidpair_equalkey> objects_type;
typedef objects_type::iterator objects_iterator;

extern objects_iterator no_object;

// --------------------------- CCoreBinFile

class ATL_NO_VTABLE CCoreBinFile : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public IDispatchImpl<ICoreStorage, &IID_ICoreStorage, &LIBID_CORELib>,
	public ISupportErrorInfoImpl<&IID_ICoreStorage>,
	public CComCoClass<CCoreBinFile, &CLSID_CoreBinFile>
{
public:
	CCoreBinFile();
	~CCoreBinFile();

BEGIN_COM_MAP(CCoreBinFile)
	COM_INTERFACE_ENTRY(ICoreStorage)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(ISupportErrorInfo)
END_COM_MAP()

DECLARE_REGISTRY_RESOURCEID(IDR_COREBINFILE)

// ------- MetaProject

public:
	CComObjPtr<ICoreMetaProject> metaproject;
	bindata metaprojectid;

	STDMETHODIMP get_MetaProject(ICoreMetaProject **p);
	STDMETHODIMP put_MetaProject(ICoreMetaProject *p);

	void OpenMetaProject();
	void CloseMetaProject() NOTHROW;

// ------- MetaObject

public:
	CComObjPtr<ICoreMetaObject> metaobject;
	metaid_type metaid;

	STDMETHODIMP get_MetaObject(ICoreMetaObject **p);
	STDMETHODIMP put_MetaObject(ICoreMetaObject *p);

	STDMETHODIMP get_MetaID(metaid_type *p);
	STDMETHODIMP put_MetaID(metaid_type p);

	void OpenMetaObject();
	void CloseMetaObject() NOTHROW;

// ------- MetaAttribute

public:
	CComObjPtr<ICoreMetaAttribute> metaattribute;
	attrid_type attrid;

	STDMETHODIMP get_MetaAttribute(ICoreMetaAttribute **p);
	STDMETHODIMP put_MetaAttribute(ICoreMetaAttribute *p);

	STDMETHODIMP get_AttrID(attrid_type *p);
	STDMETHODIMP put_AttrID(attrid_type p);

	void OpenMetaAttribute();
	void CloseMetaAttribute() NOTHROW;

// ------- Ios

public:
	ifstream ifs;
	ofstream ofs;

public:
	void read(unsigned char &a) { ifs.read((unsigned char*)&a, sizeof(unsigned char)); }
	void read(short &a) { ifs.read((unsigned char*)&a, sizeof(short)); }
	void read(int &a) { ifs.read((unsigned char*)&a, sizeof(int)); }
	void read(long &a)  { ifs.read((unsigned char*)&a, sizeof(long)); }
	void read(double &a)  { ifs.read((unsigned char*)&a, sizeof(double)); }
	void read(CComBstrObj &a);
	void read(bindata &a);

	void write(unsigned char a) { ofs.write((unsigned char*)&a, sizeof(unsigned char)); }
	void write(short a) { ofs.write((unsigned char*)&a, sizeof(short)); }
	void write(int a) { ofs.write((unsigned char*)&a, sizeof(int)); }
	void write(long a) { ofs.write((unsigned char*)&a, sizeof(long)); }
	void write(double a) { ofs.write((unsigned char*)&a, sizeof(double)); }
	void write(const CComBstrObj &a);
	void write(const bindata &a);

// ------- Attribute

public:
	typedef hash_map< BinAttrBase*, CComVariant, 
		ptr_hashfunc<BinAttrBase>, ptr_equalkey<BinAttrBase> > undos_type;
	typedef undos_type::iterator undos_iterator;

	undos_type undos;

	STDMETHODIMP get_AttributeValue(VARIANT *p);
	STDMETHODIMP put_AttributeValue(VARIANT p);

// ------- Object

public:
	objects_type objects;
	objects_iterator opened_object;

	typedef hash_map<metaid_type, objid_type> maxobjids_type;
	typedef maxobjids_type::iterator maxobjids_iterator;

	maxobjids_type maxobjids;
	slist<objects_iterator> deleted_objects;
	slist<objects_iterator> created_objects;

	struct resolve_type
	{
		objects_iterator obj;
		attrid_type attrid;
		metaobjidpair_type idpair;
	};

	typedef slist<resolve_type> resolvelist_type;
	resolvelist_type resolvelist;

	void InitMaxObjIDs();

public:
	STDMETHODIMP OpenObject(objid_type objid);
	STDMETHODIMP CreateObject(objid_type *objid);
	STDMETHODIMP CloseObject() NOTHROW;

	STDMETHODIMP DeleteObject();
	STDMETHODIMP LockObject();

// ------- Project

public:
	string filename;
	bool read_only;
	bool intrans;
	bool modified;

	bool IsOpened() const { return !filename.empty(); }
	bool InTransaction() const { return intrans; }

	void CancelProject() NOTHROW;
	void SaveProject();
	void LoadProject();

public:
	STDMETHODIMP OpenProject(BSTR connection, VARIANT_BOOL *ro_mode);
	STDMETHODIMP CreateProject(BSTR connection);
	STDMETHODIMP CloseProject( VARIANT_BOOL abort = VARIANT_FALSE) NOTHROW;
	STDMETHODIMP SaveProject(BSTR connection, VARIANT_BOOL keepoldname);
	STDMETHODIMP DeleteProject() { return E_NOTIMPL; }

	STDMETHODIMP BeginTransaction();
	STDMETHODIMP CommitTransaction();
	STDMETHODIMP AbortTransaction();

	STDMETHODIMP get_StorageType(long *p);
};

// --------------------------- BinAttr<VALTYPE_LONG>

template<>
class BinAttr<VALTYPE_LONG> : public BinAttrBase
{
public:
	BinAttr() : a(0) { }

	long a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_LONG; }
	virtual void Set(CCoreBinFile *binfile, VARIANT p)
	{ ASSERT( binfile != NULL ); binfile->modified = true; CopyTo(p, a); }

	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const { CopyTo(a, p); }
	virtual void Write(CCoreBinFile *binfile) const { binfile->write(a); }
	virtual void Read(CCoreBinFile *binfile) { binfile->read(a); }
};

// --------------------------- BinAttr<VALTYPE_REAL>

template<>
class BinAttr<VALTYPE_REAL> : public BinAttrBase
{
public:
	BinAttr() : a(0) { }

	double a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_REAL; }
	virtual void Set(CCoreBinFile *binfile, VARIANT p)
	{ ASSERT( binfile != NULL ); binfile->modified = true; CopyTo(p, a); }

	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const { CopyTo(a, p); }
	virtual void Write(CCoreBinFile *binfile) const { binfile->write(a); }
	virtual void Read(CCoreBinFile *binfile) { binfile->read(a); }
};

// --------------------------- BinAttr<VALTYPE_STRING>

template<>
class BinAttr<VALTYPE_STRING> : public BinAttrBase
{
public:
	CComBstrObj a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_STRING; }
	virtual void Set(CCoreBinFile *binfile, VARIANT p)
	{ ASSERT( binfile != NULL ); binfile->modified = true; CopyTo(p, a); }

	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const { CopyTo(a, p); }
	virtual void Write(CCoreBinFile *binfile) const { binfile->write(a); }
	virtual void Read(CCoreBinFile *binfile) { binfile->read(a); }
};

// --------------------------- BinAttr<VALTYPE_BINARY>

template<>
class BinAttr<VALTYPE_BINARY> : public BinAttrBase
{
public:
	bindata a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_BINARY; }
	virtual void Set(CCoreBinFile *binfile, VARIANT p)
	{ ASSERT( binfile != NULL ); binfile->modified = true; CopyTo(p, a); }

	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const { CopyTo(a, p); }
	virtual void Write(CCoreBinFile *binfile) const { binfile->write(a); }
	virtual void Read(CCoreBinFile *binfile) { binfile->read(a); }
};

// --------------------------- BinAttr<VALTYPE_LOCK>

template<>
class BinAttr<VALTYPE_LOCK> : public BinAttrBase
{
public:
	BinAttr() : a(0) { }

	lockval_type a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_LOCK; }
	virtual void Set(CCoreBinFile *binfile, VARIANT p) { CopyTo(p, a); }
	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const { CopyTo(a, p); }
	virtual void Write(CCoreBinFile *binfile) const { }
	virtual void Read(CCoreBinFile *binfile) { a = 0; }
};

// --------------------------- BinAttr<VALTYPE_POINTER>

template<>
class BinAttr<VALTYPE_POINTER> : public BinAttrBase
{
public:
	BinAttr() : a(no_object) { }

	objects_iterator a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_POINTER; }
	void Set(CCoreBinFile *binfile, objects_iterator a);
	virtual void Set(CCoreBinFile *binfile, VARIANT p);
	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const;
	virtual void Write(CCoreBinFile *binfile) const;
	virtual void Read(CCoreBinFile *binfile);
};

// --------------------------- BinAttr<VALTYPE_COLLECTION>

template<>
class BinAttr<VALTYPE_COLLECTION> : public BinAttrBase
{
public:
	vector<objects_iterator> a;

	virtual valtype_type GetValType() const NOTHROW { return VALTYPE_COLLECTION; }
	virtual void Set(CCoreBinFile *binfile, VARIANT p) { ASSERT(false); }
	virtual void Get(CCoreBinFile *binfile, VARIANT *p) const;
	virtual void Write(CCoreBinFile *binfile) const { }
	virtual void Read(CCoreBinFile *binfile) { }
};

#endif//MGA_COREBINFILE_H
