#pragma warning ( disable : 4786 )

#include <fstream>
#include <map>
#include <cassert>
using namespace std;

/************************SearchAndReplace class************************
Written by Jonathan Sprinkle

Purpose:  Perform a case-sensitive search and replace of strings
					in a file.

Usage:		For static calls only, modifies files on the system level,
					not the filestream level.

	Parameters:
			--- const map<string,string> &_search_replace
					contains the LHS, RHS replacement strings.  if empty, no changes
					will be made to the interior of the file (although it may still
					be renamed).

			--- const string &_oldname
					the name of the file in which to perform the search and replace

			--- const string &_newname=""
					the name of the file to be created.  if left blank, it is
					the same as the old file name.
					
			--- bool deleteOld=true
				  if true, the old file will be deleted, else it will be kept

**********************************************************************/

class SearchAndReplace
{
public:
	static void replace( const map<string,string> &_search_replace, const string &_oldname, const string &_newname="", bool deleteOld=true )
	{
		string _actualnewname;
		if( (_newname.length( ) == 0) || (_newname == _oldname))
		{ 
			_actualnewname = "__" + _oldname;
			assert( deleteOld );
		}
		else
		{
			_actualnewname = _newname;
		}
		std::ifstream _old( _oldname.data( ) );
		std::ofstream _new( _actualnewname.data( ) );
		if( _old.good( ) && _new.good( ) )
		{
			while( !_old.eof( ) && _old.good( ) )
			{
				string line;
				getline( _old, line );
				map<string,string>::const_iterator _it = _search_replace.begin( );
				for( ; _it != _search_replace.end( ); _it++ )
				{
					line = replace( line, (*_it).first, (*_it).second );
				}
				_new << line << endl;
			}
		}
		_old.close( );
		_new.close( );
		if( _newname != _actualnewname ) // deleteOld will be true...
		{
			assert( deleteOld );
			DeleteFile( _oldname.data( ) );
			MoveFile( _actualnewname.data( ), _oldname.data( ) );
		}
		else if( deleteOld )
		{
			DeleteFile( _oldname.data( ) );
		}
		return;
	}
protected:
	static string replace( const string str, const string &_search, const string &_replace )
	{
		string result=str;
		int length = _search.length( );
		int replaceLoc = result.find( _search );
		if( replaceLoc != -1 )
		{
			result.replace( replaceLoc, length, _replace );
		}
		return result;
	}
};

// end class SearchAndReplace

/************************RenameComponent class*************************
Written by Jonathan Sprinkle

Purpose:	To transform a standard GME Interpreter compiled with visual
					studio 6.0 (either BON or COM based) from the standard
					component name to the customized name of the component.  This
					allows interpreter writers to create a DLL that is the name
					of their component, as well as including more than one
					component in a workspace/solution.

					Files modified:
					- *.dsp --> becomes [componentName]originalName.dsp
									--> changes all calls to the component identity
											to match that of 'componentName' instead of 'Component'
					- *.def --> sets file access rights to normal
									--> changes the name of the output .DLL
									--> resets file access rights to readonly
					- *.dsw --> changes the name of the .DSP file included, and
											the name of the component it represents to reflect
											the wishes of the interpreter writer

**********************************************************************/
class RenameComponent
{
public:
	RenameComponent( string _componentName, string _dspFileName, string _oldComponentName="Component" )
		: componentName( _componentName ), oldComponentName( _oldComponentName ),
			componentizedFilename( _componentName + _dspFileName ),
			dspFileName( _dspFileName )
	{
		if( oldComponentName.length( ) > 0 && componentizedFilename.find( oldComponentName ) != -1 )
		{
			componentizedFilename.replace( componentizedFilename.find( oldComponentName ), oldComponentName.length( ), "" );
		}
		if( oldComponentName == "" ) oldComponentName = "Component";
		dsp2dsw["Component.dsp"]		= "Component.dsw";
		dsp2dsw["BONComponent.dsp"]	= "BONComponent.dsw";
		dsp2dsw["BON2Component.dsp"]	= "BON2Component.dsw";
		if( oldComponentName.compare( "Component" ) != 0 )
		{
			if( _dspFileName.find( "BON2" ) != -1 )
			{
				dsp2dsw[oldComponentName + "BON2Component.dsp"] = "BON2Component.dsw";
			}
			else if( _dspFileName.find( "BON" ) != -1 )
			{
				dsp2dsw[oldComponentName + "BONComponent.dsp"] = "BONComponent.dsw";
			}
			else
			{
				dsp2dsw[oldComponentName + "Component.dsp"] = "Component.dsw";
			}
		}
		doDsp( );
		doDef( );
		doDsw( );
	}
	string getComponentizedFilename( ) { return componentizedFilename; }
	string getComponentName( ) { return componentName; }
	map<string, string> dsp2dsw;

protected:
	string componentName;
	string componentizedFilename;
	string dspFileName;
	string oldComponentName;
	void doDsp( )
	{
		map<string,string> _s_r;
		_s_r["\"" + oldComponentName + " "] = "\"" + componentName + " ";
		_s_r["\"" + oldComponentName + "\""] = "\"" + componentName + "\"";
		_s_r[oldComponentName + ".dll"] = componentName + ".dll";
		SearchAndReplace::replace( _s_r, dspFileName, componentizedFilename );
	}
	void doDef( )
	{
		map<string,string> _s_r;
		_s_r[oldComponentName + ".DLL"] = componentName + ".DLL";
		SetFileAttributes( "component.def", FILE_ATTRIBUTE_NORMAL );
		SearchAndReplace::replace( _s_r, "component.def" );
		SetFileAttributes( "component.def", FILE_ATTRIBUTE_READONLY );
	}
	void doDsw( )
	{
		assert( dsp2dsw.find(dspFileName) != dsp2dsw.end( ) );
		string dswName = dsp2dsw[dspFileName];
		map<string,string> _s_r;
		_s_r["\"" + oldComponentName + "\""] = "\"" + componentName + "\"";
		_s_r[dspFileName] = componentizedFilename;
		SearchAndReplace::replace( _s_r, dswName );
	}

};

// end class RenameComponent
